{bridge}                = require "./../Bridge"
{ViewerTools}           = require "./../tools/ViewerTools"
{log}                   = require "./../tools/Log"
{HighlightComponent}    = require "./../highlight/HighlightComponent"
{EditStateController}   = require "./../edit/EditStateController"
{Transform}             = require "./../tools/Transform"
require "./Styles"

class exports.StateController

    constructor: (@deviceContext) ->

        log "Setting up state controller", @deviceContext

        @selecting = false
        @clientLayers = []
        @eventLayers = []

        @directSelectionAllowed = false

        @state = {
            mode: "Default"
            codeState: "OK"
            selecting: false
            highlightingTarget: {}
            editingTarget: {}
        }

        # Layer selection highlighting
        @highlighter = new HighlightComponent()

        # Layer editing
        @editor = new EditStateController()

        # Listen for state changes:
        bridge.on "studio:state", (state) =>

            log "StateController:state handler;", state

            switch state.mode
                when "Editing"      then @stopSelecting()
                when "Selecting"    then @startSelecting()  # Direct selection is disabled for now.
                when "Highlighting" then @stopSelecting()
                else                     @stopSelecting()

            @renderState state

        # Listen for layer changes from the properties panel:
        bridge.on "propertiesPanel:propertyChanged", (data) =>

            layerHash = @state.editingTarget?.owner
            return if not layerHash?

            layer = ViewerTools.getLayerByHash(layerHash)
            return if not layer?

            coercedValue =
                if not data.value? or data.value is ""
                    defaultValue = layer.states._states.default[data.name]
                    if not defaultValue?
                        if data.type is "number" or data.type is "@Number"
                            defaultValue = 0
                    defaultValue
                else if data.type is "number" or data.type is "@Number"
                    parseFloat(data.value)
                else
                    data.value.replace?(/^"(.+(?="$))"$/, '$1')

            layer[data.name] = coercedValue
            @editor.editLayer layer
            layer[data.name] = coercedValue for layer in @editor.relatedLayers if @editor.relatedLayers?

        # Listen for the device context being reset, and clean up lingering
        # edit targets:
        @deviceContext.on "reset", =>
            log "device context reset - StateController"

            if @state?.highlightingTarget or @state?.editingTarget
                @state?.editingTarget = null
                @state?.highlightingTarget = null
                @renderState @state

        # We may be reset while frozen. This leads to trouble when we are trying
        # to unfreeze, because listeners might be put back on layers that longer
        # exists. Runtime therefore sends out a "willReset" via the context, before
        # calling "reset" on the context. We use this to unfreeze (exit selectmode)
        # prior to that:
        @deviceContext.on "willReset", =>
            @stopSelecting()

    # Render
    #

    renderState: (newState) ->

        highlightingLayerId = newState.highlightingTarget?.runtimeId
        currentHighlightingLayerId = @state.highlightingTarget?.runtimeId

        editingTargetLayerHash = newState.editingTarget?.owner
        editingTargetLayerState = newState.editingTarget?.data?.state || "default"
        editingTargetFrozenMembers = newState.editingTarget?.frozenMembers

        currentEditingTargetLayerHash = @state.editingTarget?.owner

        # Switch in the 'arrow' cursor when we're doing direct manipulation:
        if newState.selecting or editingTargetLayerHash? or highlightingLayerId?
            document.body.classList.add("editMode")
        else
            document.body.classList.remove("editMode")

        # Apply the correct highlighting:
        if highlightingLayerId isnt currentHighlightingLayerId
            if highlightingLayerId? and not editingTargetLayerHash?
                @highlighter.highlight(
                    ViewerTools.getLayerById(highlightingLayerId),
                    not newState.editingTarget?.owner?
                )
            else
                @highlighter.unhighlight()
        else
            @highlighter.unhighlight()

        # Apply the correct editing chrome:
        if editingTargetLayerHash isnt currentEditingTargetLayerHash

            layer = ViewerTools.getLayerByHash(editingTargetLayerHash)
            if layer && layer.states.currentState != editingTargetLayerState
                # The main layer is already switched, as the property getter does so:
                # layer.states.switchInstant(editingTargetLayerState)
                for relatedLayer in ViewerTools.getRelatedLayersByHash(editingTargetLayerHash)
                    relatedLayer.states.switchInstant(editingTargetLayerState)

            @editor.propertyLock.frozenMembers = null
            @editor.propertyLock.frozenMembers = editingTargetFrozenMembers if editingTargetFrozenMembers?
            @editor.editHash(editingTargetLayerHash)

        @state = newState

    # Context wide
    #

    handleContextMouseDown: (event) =>

        return unless @state.selecting

        event.preventDefault()
        event.stopImmediatePropagation()

        target = event.target
        while target && !target.classList.contains("framerLayer")
            target = target.parent

        if target? and (layer = target.__framerInstance)?

            @mouseDownLayer = layer

            document.body.addEventListener "mouseup", @handleContextMouseUp, true

            @mouseDownFrame = Transform.getLayerScreenFrame(layer)
            @mouseDownX = event.screenX
            @mouseDownY = event.screenY

            if @directSelectionAllowed
                bridge.send "framerView:layerClick",
                    hash: layer.__framerInstanceInfo?.hash
                    state: layer.states.current

    handleContextMouseUp: (event) =>

        return unless @state.mode is "Selecting"

        document.body.removeEventListener "mouseup", @handleContextMouseUp, true

        event.preventDefault()
        event.stopImmediatePropagation()

        @mouseDownLayer = null if @mouseDownLayer?

    handleContextMove: (event) =>

        return unless @state.selecting

        # When the left mouse button is down, then we're dragging: handle this as
        # a dragging event:
        return if event.which is 1

        # Mouse move is triggered on CMD up. This means we're escaping
        # selection mode but that the FS state hasn't updated yet. If
        # that's the case, ignore this event:
        return unless event.metaKey

        # Ignore mouse over's when in 'selecting' mode, and direct select is "off":
        # return if not @directSelectionAllowed and @state.mode is "Editing"

        event.preventDefault()
        event.stopImmediatePropagation()

        # Traverse the DOM tree from the target node up to the first div that is
        # a "framerLayer":
        target = event.target
        while target && !target.classList.contains("framerLayer")
            target = target.parent

        if (target? and
            (layer = target.__framerInstance)?
        )
            # if layer.id isnt @state.highlightingTarget.runtimeId
            bridge.send "framerView:layerHighlighted",
                id: layer.id
                hash: layer.__framerInstanceInfo?.hash
                state: layer.states.current
        else
            bridge.send "framerView:unhighlightLayer"


    # Select Mode
    #

    startSelecting: ->
        return if @selecting or not @editor.mouse?
        @selecting = true

        context = @deviceContext

        @clientLayers = context._layers.concat() # copy

        for layer in @clientLayers
            if layer.ignoreEvents
              layer.ignoreEvents = false
            else
              @eventLayers.push(layer)
            layer._element.__framerInstance = layer

        document.body.addEventListener "mousemove", @handleContextMove, true
        document.body.addEventListener "mousedown", @handleContextMouseDown, true

    stopSelecting: ->
        return unless @selecting

        context = @deviceContext

        document.body.removeEventListener "mousedown", @handleContextMouseDown, true
        document.body.removeEventListener "mousemove", @handleContextMove, true

        for layer in @clientLayers
            if layer not in @eventLayers
              layer.ignoreEvents = true
            delete layer._element.__framerInstance

        @clientLayers = []
        @eventLayers = []
        @highlighter.unhighlight()
        @selecting = false

    # Utils
    #

    log: (args...) ->
        # console.log args...
